﻿/*
 **********************************************************************
 * Date of Creation : 20111028                                        *
 *           Author : Gianni.Cavallotto@ors.it                        *
 *      Description : derivation of UltraGrid with support features   *
 *                    for pagination                                  *
 ********************************************************************** 
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Infragistics.Win.UltraWinGrid;
using Infragistics.Win.UltraWinDataSource;
using System.Data;
using System.Windows.Forms;
using System.Collections;
using System.ComponentModel;
using Infragistics.Win.Misc;
using System.IO;
using Infragistics.Shared;
using PaginatedUltraWinGrid;

namespace Ors.Rams.Workstation.Common.InfragisticsIntegration
{
    public class UltraGridPaginated : UltraGrid
    {

        #region Consts

        private const int MinimumChunkSize = 40; // Data chunk minimum size; has to be an even number
        private const int InitialChunkSize = 100; // Initial default data chunk size
        private const int ChunkSizeVariation = 20; // Data chunk autosize variation step 
        private const long OptimalChunkWait = 150; // number of milliseconds for optimal data chunk wait time 
        private const int MaxGroupingWait = 5000; // number of millisends acceptable to wait for grouping without warn the user
        private const int BackGroundRetrievalInterval = 5500; // in milliseconds; interval for automatic background data retrieval
        private const double TimeOutInterval = 30000.0; // number of milliseconds after which data connection will be considered lost

        private const int WM_LBUTTONDOWN = 0x0201;
        private const int WM_LBUTTONUP = 0x0202;
        private const int WM_LBUTTONDBLCLK = 0x0203;
        private const int WM_RBUTTONDOWN = 0x0204;
        private const int WM_RBUTTONUP = 0x0205;
        private const int WM_RBUTTONDBLCLK = 0x0206;
        private const int WM_MBUTTONDOWN = 0x0207;
        private const int WM_MBUTTONUP = 0x0208;
        private const int WM_MBUTTONDBLCLK = 0x0209;
        private const int WM_MOUSEWHEEL = 0x020A;

        #endregion


        #region Events

        /// <summary>
        /// Fired any time grid need a new chunk of data.
        /// Listeners have to provide data with the public method AddChunk . 
        /// </summary>
        public event EventHandler<ChunkRequestEventArgs> ChunkRequesting;
        private void chunkRequesting(ChunkRequestEventArgs args)
        {
            if (this.ChunkRequesting != null) {
                this.timeOut.Stop();
                this.timeOut.Start();
                this.ChunkRequesting(this, args);
            }
        }

        /// <summary>
        /// Fired any time a column distinct values usable to filter has to be shown.
        /// Listeners have to provide data with the public method SetFilterValues .  
        /// </summary>
        public event EventHandler<FilterValuesEventArgs> FilterValuesRequesting;
        private void filterValuesRequesting(IEnumerable<string> columnsKeys)
        {
            if (this.FilterValuesRequesting == null)
                return;

            this.FilterValuesRequesting(this, new FilterValuesEventArgs(columnsKeys, this.DisplayLayout.Bands[0].ColumnFilters));
        }

        #endregion


        #region Fields

        private ArrayList dataRepository; // data cache
        private UltraDataSource ultraDataSource = new UltraDataSource(); // UltraGrid binding object for data OnDemand.
        private DataTable dataStructureTable; // grid structure repository.
        private bool dataProcessing = false; // flag to know if data loading engine is busy or not.
        private int requestingIndex = 0; // index of the first record of the requested data chunk.
        private System.Diagnostics.Stopwatch dataPumpWatch = new System.Diagnostics.Stopwatch(); // used to measure data acquiring times.
        private Dictionary<int, Queue<long>> measurements = new Dictionary<int, Queue<long>>(); // measurement repository.
        private UltraGridRow lastActiveRow; // selfexplaned
        private bool backScrolling = false; // selfexplaned
        private Dictionary<string, List<object>> filterDistinctValues = new Dictionary<string, List<object>>(); // selfexplaned.
        private bool allDataIsLoaded = false; // flag to know if entire dataset is already loaded (so pagination will no longer be usefull).
        private bool allDataIsLoading = false; // flag to know if a load of entire dataset has been requested and is in progress.
        private MemoryStream displayStream; // Temporary repository of the grid display layout.
        private Timer backGroundTimer = new Timer(); // Times the automatic data retrieval.
        private BackgroundWorker backRetrievalWorker = new BackgroundWorker(); // Handles the automatic data retrieval.
        private System.Timers.Timer timeOut = new System.Timers.Timer(); // Times out the data connection

        #endregion


        #region Properties

            private int chunkSize = InitialChunkSize;
            /// <summary>
            /// Data chunk size expressed in number of rows. To enable/disable autosizing set the boolean property "AutoSizeChunk".
            /// Autosizing is enabled by default.
            /// </summary>
            [DefaultValueAttribute(InitialChunkSize)]
            public int ChunkSize
            {
                get
                {
                    return this.chunkSize;
                }
                set
                {
                    if (value >= MinimumChunkSize)
                        this.chunkSize = value;
                }
            }

            private bool autoChunk = true;
            /// <summary>
            /// Enable / Disable the data chunk automatic dimensioning.
            /// </summary>
            [DefaultValueAttribute(true)]
            public bool AutoSizeChunk
            {
                get { return this.autoChunk; }
                set { this.autoChunk = value; }
            }

            /// <summary>
            /// Built-in adapter to facilitate sort status export versus common DataSource repository types.
            /// If your data repository is not provided please consider to implement your own adaptor.
            /// </summary>
            public UltraGridSortAdapter SortAdapter { get; internal set; }
            /// <summary>
            /// Built-in adapter to facilitate filter status export versus common DataSource repository types.
            /// If your data repository is not provided please consider to implement your own adaptor.
            /// </summary>
            public UltraGridFilterAdapter FilterAdapter { get; internal set; }

            /// <summary>
            /// Get / Set the number of rows the grid has to be dimensionate. 
            /// When filtering you would variate this value.
            /// </summary>
            public int? TotalRows { get; set; }

        #endregion


        public UltraGridPaginated()
        {
            this.SortAdapter = new UltraGridSortAdapter();
            this.FilterAdapter = new UltraGridFilterAdapter();

            this.backGroundTimer.Interval = BackGroundRetrievalInterval;
            this.backGroundTimer.Start();

            this.timeOut.Interval = TimeOutInterval;

            this.subscribeEvents();
        }


        #region Event Handles

        void timeOut_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            MessageBox.Show(UltraGridPaginatedResources.CONNECTION_TIMEOUT, UltraGridPaginatedResources.DATA_FAILURE
                            , MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1, MessageBoxOptions.DefaultDesktopOnly);
            this.timeOut.Stop();
        }

        void backRetrievalWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            if (this.dataRepository != null) {
                int index = this.dataRepository.IndexOf(null);
                if (index < 0) {
                    // no null records
                    if (!this.gridIsFiltered()) {
                        if (this.TotalRows != null && this.dataRepository.Count == this.TotalRows.Value) {
                            // if no filters are set and all records have been loaded 
                            this.allDataIsLoaded = true;
                            this.backGroundTimer.Stop();
                        } else if (this.dataRepository.Count < this.ultraDataSource.Rows.Count) {
                            // request next data chunk to be queued
                            this.requestChunk(this.dataRepository.Count, this.DisplayLayout.Bands[0].ColumnFilters, this.DisplayLayout.Bands[0].SortedColumns);
                        }
                    }
                } else {
                    this.requestChunk(index, this.DisplayLayout.Bands[0].ColumnFilters, this.DisplayLayout.Bands[0].SortedColumns);
                }
            }
        }

        void backGroundTimer_Tick(object sender, EventArgs e)
        {
            if (!this.backRetrievalWorker.IsBusy) {
                this.backRetrievalWorker.RunWorkerAsync();
            }
        }

        void UltraGridPaginated_MouseDown(object sender, MouseEventArgs e)
        {
            this.backScrolling = false;
        }

        void UltraGridPaginated_BeforeRowFilterDropDownPopulate(object sender, BeforeRowFilterDropDownPopulateEventArgs e)
        {
            foreach (object item in this.filterDistinctValues[e.Column.Key])
                e.ValueList.ValueListItems.Add(item);
            e.Handled = true;
        }

        void UltraGridPaginated_PropertyChanged(object sender, Infragistics.Win.PropertyChangedEventArgs e)
        {
            // If filter change
            PropChangeInfo columnFilterChangeInfo = e.ChangeInfo.FindPropId(Infragistics.Win.UltraWinGrid.PropertyIds.ColumnFilters);
            if (columnFilterChangeInfo != null) {
                if (!this.allDataIsLoaded) {
                    this.dataRepository = new ArrayList();

                    List<string> list = new List<string>();
                    foreach (UltraGridColumn col in this.DisplayLayout.Bands[0].Columns) {
                        list.Add(col.Key);
                    }

                    this.filterValuesRequesting(list);
                }
            }
            // If DataSource change
            //PropChangeInfo dataSourceChangeInfo = e.ChangeInfo.FindPropId(Infragistics.Win.UltraWinGrid.PropertyIds.DataSource);
            //if (dataSourceChangeInfo != null) {
            //}
        }

        void UltraGridPaginated_BeforeSortChange(object sender, BeforeSortChangeEventArgs e)
        {
            if (this.dataRepository != null && !this.allDataIsLoaded) {
                this.dataRepository = new ArrayList();

                bool isGrouping = false;
                foreach (UltraGridColumn col in e.SortedColumns) {
                    if (col.IsGroupByColumn) {
                        isGrouping = true;
                        break;
                    }
                }
                if (isGrouping) {

                    // approx needed time calculation
                    int notNulls = 0;
                    foreach (var item in this.dataRepository) {
                        if (item != null)
                            notNulls++;
                    }
                    long maxAcceptableMissingChunks = MaxGroupingWait / OptimalChunkWait;
                    if ((this.TotalRows - notNulls) > this.ChunkSize * maxAcceptableMissingChunks) {
                        DialogResult dR = MessageBox.Show(String.Format(UltraGridPaginatedResources.GROUPING_PROCEED, (((this.TotalRows - notNulls) / this.ChunkSize) * OptimalChunkWait / 1000).ToString())
                                                            , UltraGridPaginatedResources.PROCEED
                                                            , MessageBoxButtons.YesNo
                                                            , MessageBoxIcon.Question
                                                            , MessageBoxDefaultButton.Button2
                                                            , MessageBoxOptions.DefaultDesktopOnly);
                        if (dR != DialogResult.Yes) {
                            e.Cancel = true;
                        }
                    }

                }
            }
        }

        void UltraGridPaginated_AfterSortChange(object sender, BandEventArgs e)
        {
            if (this.dataRepository != null && !this.allDataIsLoaded) {

                bool isGrouping = false;
                foreach (UltraGridColumn col in e.Band.SortedColumns) {
                    if (col.IsGroupByColumn) {
                        isGrouping = true;
                        break;
                    }
                }
                if (isGrouping) {
                    this.loadAllData();
                }
            }
        }

        void UltraGridPaginated_MouseWheel(object sender, MouseEventArgs e)
        {
            if (e.Delta > 0)
                this.backScrolling = true;
            else
                this.backScrolling = false;
        }

        void UltraGridPaginated_AfterRowActivate(object sender, EventArgs e)
        {
            if (this.dataRepository != null) {
                if (this.lastActiveRow != null) {
                    this.backScrolling = false;
                    int k = this.ActiveRow.Index > this.lastActiveRow.Index ? this.ChunkSize / 2 : -(this.ChunkSize / 2);
                    int newIndex = this.ActiveRow.Index + k;
                    if (!(newIndex < 0)) {
                        if (this.dataRepository.Count < newIndex + 1
                            || this.dataRepository[newIndex] == null) {
                            if (k > 0) {
                                this.requestChunk(this.ActiveRow.Index + 1, this.DisplayLayout.Bands[0].ColumnFilters, this.DisplayLayout.Bands[0].SortedColumns);
                            } else {
                                this.requestChunk(this.ActiveRow.Index - this.ChunkSize - 1, this.DisplayLayout.Bands[0].ColumnFilters, this.DisplayLayout.Bands[0].SortedColumns);
                            }
                        }
                    }
                }
                this.lastActiveRow = this.ActiveRow;
            }
        }

        void ultraDataSource_CellDataRequested(object sender, CellDataRequestedEventArgs e)
        {
            DataRow dR = null;

            if (e.Row.Index < this.dataRepository.Count)
                dR = this.dataRepository[e.Row.Index] as DataRow;

            if (dR != null) {
                if (e.Column.Key != "RowIndex") {
                    e.Data = dR[e.Column.Key];
                }
            } else {
                if (this.backScrolling) {
                    int index = e.Row.Index - this.ChunkSize / 2;
                    if (index < 0)
                        index = 0;
                    this.requestChunk(index, this.DisplayLayout.Bands[0].ColumnFilters, this.DisplayLayout.Bands[0].SortedColumns);
                } else
                    this.requestChunk(e.Row.Index, this.DisplayLayout.Bands[0].ColumnFilters, this.DisplayLayout.Bands[0].SortedColumns);
                // temporary data to be shown
                if (e.Column.DataType == typeof(string))
                    e.Data = "...";
            }

            // Set CacheData to false to prevent UltraDataSource from caching the 
            // data. Default value of CacheData property is true.
            e.CacheData = false;
        }

        #endregion


        #region Overrides

        protected override void WndProc(ref Message m)
        {
            if (this.allDataIsLoading) {
                if (m.Msg > 0x0200 && m.Msg < 0x020B) {
                    //Dont dispatch message
                    return;
                }
            }

            //Dispatch as usual
            base.WndProc(ref m);
        }

        protected override void Dispose(bool disposing)
        {
            this.unsubscribeEvents();
            base.Dispose(disposing);
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// To be used to provide a data chunk.
        /// If the chunk to be provided is the first one please set TotalRows property before or consider using AddChunk(object chunk, int? totalRowCount) overload.
        /// </summary>
        /// <param name="chunk">Data chunk object</param>
        public void AddChunk(object chunk)
        {
            this.AddChunk(chunk, null);
        }

        /// <summary>
        /// To be used to provide a data chunk.
        /// Set <paramref name="totalRowCount"/> parameter to total amount of row, regardless of data chunk size and according to set filters.
        /// If total rows size is already been set and it isn't changed, <paramref name="totalRowCount"/> parameter could be set to null.
        /// </summary>
        /// <param name="chunk">Data chunk object</param>
        /// <param name="totalRowCount">Total number of rows (according to set filters)</param>
        public void AddChunk(object chunk, int? totalRowCount)
        {
            this.timeOut.Stop();

            lock (this) {
                // initialization of a shadow grid used to emulate Infragistics binding policy that will be inherited
                UltraGrid grid = new UltraGrid();
                grid.CreateControl();
                grid.BindingContext = new BindingContext();
                grid.DataSource = chunk;
                grid.DataBind();

                // If this is the absolute first data chunk got
                if (this.dataRepository == null) {

                    // ...repositories initialization
                    this.dataRepository = new ArrayList();
                    this.dataStructureTable = new DataTable();
                    this.ultraDataSource.Reset();

                    // ...ultraDataSource and data structure initialization
                    Dictionary<string, string> captionsMap = new Dictionary<string, string>();
                    foreach (UltraGridColumn col in grid.DisplayLayout.Bands[0].Columns) {
                        this.ultraDataSource.Band.Columns.Add(col.Key, col.DataType);
                        this.dataStructureTable.Columns.Add(col.Key, col.DataType);
                        captionsMap.Add(col.Key, col.Header.Caption);
                    }

                    if (this.TotalRows == null && totalRowCount != null)
                        this.TotalRows = totalRowCount;
                    this.DataSource = this.ultraDataSource;
                    this.DisplayLayout.LoadStyle = LoadStyle.LoadOnDemand;

                    // I decided to hide "Match" and "Does not Match" filter options
                    // because too much expensive to handle regular-expression filter values and not so worthy
                    foreach (UltraGridColumn uC in this.DisplayLayout.Bands[0].Columns) {
                        if (uC.DataType == typeof(string)) {
                            uC.FilterOperatorDropDownItems = FilterOperatorDropDownItems.All ^ (FilterOperatorDropDownItems.DoesNotMatch | FilterOperatorDropDownItems.Match);
                        }
                    }

                    // column captions set
                    foreach (UltraGridColumn c in this.DisplayLayout.Bands[0].Columns)
                        c.Header.Caption = captionsMap[c.Key];

                    // filters distinct values request
                    this.filterValuesRequesting(captionsMap.Keys);

                } else {
                    // consistency data chunk structure check
                    foreach (UltraGridColumn column in grid.DisplayLayout.Bands[0].Columns)
                        if (!this.ultraDataSource.Band.Columns.Exists(column.Key))
                            throw new FormatException("Provided data do not match with the current visualized data format: impossible to find column \"[" + column.Key + "]" + column.Header.Caption + "\".");
                }

                // Fill gap with null values
                for (int j = this.dataRepository.Count; j < this.requestingIndex; j++)
                    this.dataRepository.Add(null);

                // Load data  
                for (int k = this.requestingIndex; k < this.requestingIndex + grid.Rows.Count; k++) {
                    int k_Resolved = this.backScrolling ? (this.requestingIndex + grid.Rows.Count) - (k - this.requestingIndex) - 1 : k;
                    DataRow dR = this.dataStructureTable.NewRow();
                    foreach (UltraGridColumn column in grid.DisplayLayout.Bands[0].Columns) {
                        dR[column.Key] = grid.Rows[k_Resolved - this.requestingIndex].Cells[column].Value;
                    }
                    if (k < this.dataRepository.Count)
                        this.dataRepository[k_Resolved] = dR;
                    else
                        this.dataRepository.Add(dR);
                }

                if (dataPumpWatch.IsRunning)
                    this.chunkSizing(this.dataPumpWatch.ElapsedMilliseconds, grid.Rows.Count);

                grid.Dispose();


                if (this.allDataIsLoading) {
                    this.allDataIsLoaded = true;
                    this.allDataIsLoading = false;

                    this.UseWaitCursor = false;
                    this.Cursor = this.Cursor; // necessary to refresh the cursor... :(

                    this.Rows.Refresh(RefreshRow.ReloadData);
                    this.DisplayLayout.Load(this.displayStream);

                    // Restore CellDataRequest event subscription
                    this.ultraDataSource.CellDataRequested += new CellDataRequestedEventHandler(ultraDataSource_CellDataRequested);
                }

                if (totalRowCount != null) {
                    if (totalRowCount.Value != this.ultraDataSource.Rows.Count)
                        this.ultraDataSource.Rows.SetCount(totalRowCount.Value);
                } else if (this.TotalRows != null && this.TotalRows != this.ultraDataSource.Rows.Count) {
                    this.ultraDataSource.Rows.SetCount(this.TotalRows.Value);
                }

                // Refresh
                this.Invalidate(this.ClientRectangle);
                this.Rows.Refresh(RefreshRow.ReloadData);
                if (!this.gridIsFiltered())
                    try {
                        this.Refresh();
                    } catch { }

                this.dataProcessing = false;
                this.backScrolling = false;
            }

        }


        /// <summary>
        /// To be use to provide available distict values of a column.
        /// </summary>
        /// <param name="columnKey">The key of the column</param>
        /// <param name="values">List of distincted values of the column</param>
        public void SetFilterValues(string columnKey, IEnumerable<object> values)
        {
            BackgroundWorker worker = new BackgroundWorker();
            worker.DoWork += new DoWorkEventHandler(worker_DoWork);
            worker.RunWorkerAsync(new object[] { columnKey, values });
            worker.Dispose();
        }
        void worker_DoWork(object sender, DoWorkEventArgs e)
        {
            string columnKey = ((e.Argument as object[]))[0].ToString();
            IEnumerable<object> values = ((e.Argument as object[]))[1] as IEnumerable<object>;
            lock (this.filterDistinctValues) {
                if (this.filterDistinctValues.ContainsKey(columnKey))
                    this.filterDistinctValues[columnKey].Clear();
                else
                    this.filterDistinctValues.Add(columnKey, new List<object>());

                foreach (object item in values) {
                    this.filterDistinctValues[columnKey].Add(item);
                }
            }
        }

        #endregion


        #region Private Methods

        private object DefaultValue(Type myType)
        {
            if (!myType.IsValueType)
                return null;
            else
                return Activator.CreateInstance(myType);
        }

        private void loadAllData()
        {
            this.UseWaitCursor = true;

            // flags initialization
            this.allDataIsLoading = true;
            this.backScrolling = false;

            // Stop timers
            this.backGroundTimer.Stop();
            this.timeOut.Stop();

            // detach CellDataRequest event subscription to avoid useless overhead
            this.ultraDataSource.CellDataRequested -= this.ultraDataSource_CellDataRequested;

            // layout saving
            this.displayStream = new MemoryStream();
            this.DisplayLayout.Save(this.displayStream);
            this.displayStream.Position = 0;

            // Data request
            this.ChunkSize = int.MaxValue;
            this.requestChunk(0, null, null);
        }

        private void requestChunk(int index, ColumnFiltersCollection filter, SortedColumnsCollection sort)
        {
            if (!this.dataProcessing) {
                this.dataProcessing = true;
                this.dataPumpWatch.Reset();
                this.dataPumpWatch.Start();
                this.requestingIndex = index;
                this.chunkRequesting(new ChunkRequestEventArgs(this.requestingIndex, this.ChunkSize, filter, sort));
            }
        }

        private void chunkSizing(long elapsed, int chunkSize)
        {
            if (!this.measurements.ContainsKey(chunkSize)) {
                this.measurements.Add(chunkSize, new Queue<long>());
            }
            this.measurements[chunkSize].Enqueue(elapsed);
            if (this.measurements[chunkSize].Count > 10)
                this.measurements[chunkSize].Dequeue();
            if (this.autoChunk) {
                Dictionary<int, long> averages = new Dictionary<int, long>();
                foreach (var size in this.measurements.Keys) {
                    averages.Add(size, 0);
                    int counter = 0;
                    foreach (long time in this.measurements[size]) {
                        averages[size] += time;
                        counter++;
                    }
                    averages[size] = averages[size] / counter;
                }

                if (averages.ContainsKey(this.ChunkSize)) {
                    if (averages[this.ChunkSize] < OptimalChunkWait) {
                        this.chunkSize += ChunkSizeVariation;
                    } else {
                        this.chunkSize -= ChunkSizeVariation;
                        if (this.chunkSize < MinimumChunkSize)
                            this.chunkSize = MinimumChunkSize;
                    }
                }
            }
        }

        private bool gridIsFiltered()
        {
            bool retVal = false;
            foreach (UltraGridBand band in this.DisplayLayout.Bands) {
                foreach (UltraGridColumn column in band.Columns) {
                    if (band.ColumnFilters[column].FilterConditions.Count > 0) {
                        retVal = true;
                        break;
                    }
                }
                if (retVal)
                    break;
            }
            return retVal;
        }

        private void subscribeEvents()
        {
            this.ultraDataSource.CellDataRequested += new CellDataRequestedEventHandler(ultraDataSource_CellDataRequested);
            this.AfterRowActivate += new EventHandler(UltraGridPaginated_AfterRowActivate);
            this.MouseWheel += new MouseEventHandler(UltraGridPaginated_MouseWheel);
            this.MouseDown += new MouseEventHandler(UltraGridPaginated_MouseDown);

            this.BeforeSortChange += new BeforeSortChangeEventHandler(UltraGridPaginated_BeforeSortChange);
            this.AfterSortChange += new BandEventHandler(UltraGridPaginated_AfterSortChange);
            //this.AfterRowFilterChanged += new AfterRowFilterChangedEventHandler(UltraGridPaginated_AfterRowFilterChanged);
            this.PropertyChanged += new Infragistics.Win.PropertyChangedEventHandler(UltraGridPaginated_PropertyChanged);
            this.BeforeRowFilterDropDownPopulate += new BeforeRowFilterDropDownPopulateEventHandler(UltraGridPaginated_BeforeRowFilterDropDownPopulate);
            this.backGroundTimer.Tick += new EventHandler(backGroundTimer_Tick);
            backRetrievalWorker.DoWork += new DoWorkEventHandler(backRetrievalWorker_DoWork);
            this.timeOut.Elapsed += new System.Timers.ElapsedEventHandler(timeOut_Elapsed);
        }

        private void unsubscribeEvents()
        {
            this.ultraDataSource.CellDataRequested -= this.ultraDataSource_CellDataRequested;
            this.AfterRowActivate -= this.UltraGridPaginated_AfterRowActivate;
            this.MouseWheel -= this.UltraGridPaginated_MouseWheel;
            this.MouseDown -= this.UltraGridPaginated_MouseDown;

            this.BeforeSortChange -= this.UltraGridPaginated_BeforeSortChange;
            this.AfterSortChange -= this.UltraGridPaginated_AfterSortChange;
            this.PropertyChanged -= this.UltraGridPaginated_PropertyChanged;
            this.BeforeRowFilterDropDownPopulate -= this.UltraGridPaginated_BeforeRowFilterDropDownPopulate;
            this.backGroundTimer.Tick -= this.backGroundTimer_Tick;
            backRetrievalWorker.DoWork -= this.backRetrievalWorker_DoWork;
            this.timeOut.Elapsed -= this.timeOut_Elapsed;
        }

        #endregion

    }


    public class ChunkRequestEventArgs : EventArgs
    {
        public int LastIndex { get; set; }
        public int ChunkSize { get; set; }
        public ColumnFiltersCollection Filter { get; set; }
        public SortedColumnsCollection Sort { get; set; }

        public ChunkRequestEventArgs(int lastIndex, int chunkSize, ColumnFiltersCollection filter, SortedColumnsCollection sort)
        {
            this.LastIndex = lastIndex;
            this.ChunkSize = chunkSize;
            this.Sort = sort;
            this.Filter = filter;
        }
    }

    public class FilterValuesEventArgs : EventArgs
    {
        public IEnumerable<string> ColumnsKeys { get; set; }
        public ColumnFiltersCollection Filter { get; set; }

        public FilterValuesEventArgs(IEnumerable<string> columnsKeys, ColumnFiltersCollection filter)
        {
            this.ColumnsKeys = columnsKeys;
            this.Filter = filter;
        }
    }


    public class UltraGridSortAdapter
    {
        public object TranslateSort<T>(SortedColumnsCollection sort)
        {
            object retVal = null;

            if (typeof(T) == typeof(DataTable)) {
                string output = String.Empty;
                if (sort != null) {
                    foreach (UltraGridColumn column in sort) {
                        if (output != String.Empty)
                            output += ",";
                        output += "[" + column.Key + "] ";
                        if (column.SortIndicator == SortIndicator.Ascending)
                            output += "ASC";
                        if (column.SortIndicator == SortIndicator.Descending)
                            output += "DESC";
                    }
                }
                retVal = output;
            }

            return retVal;
        }

    }

    public class UltraGridFilterAdapter
    {

        public object TranslateFilter<T>(ColumnFiltersCollection filter)
        {
            object retVal = null;

            if (typeof(T) == typeof(DataTable)) {
                string output = String.Empty;
                string partial = String.Empty;

                if (filter != null) {
                    bool link = false;
                    foreach (ColumnFilter colFilter in filter) {
                        if (partial != String.Empty) {
                            link = true;
                            partial = String.Empty;
                        }

                        foreach (FilterCondition cond in colFilter.FilterConditions) {
                            partial += "[" + cond.Column.Key + "] ";
                            partial += TranslateSintax(cond);
                        }

                        if (partial != String.Empty && link) {
                            switch (colFilter.LogicalOperator) {
                                case FilterLogicalOperator.And:
                                    output += " AND ";
                                    break;
                                case FilterLogicalOperator.Or:
                                    output += " OR ";
                                    break;
                                default:
                                    //output += " AND ";
                                    break;
                            }
                            link = false;
                        }

                        output += partial;

                    }
                }
                retVal = output;
            }

            return retVal;
        }


        private string TranslateSintax(FilterCondition cond)
        {
            string output = String.Empty;
            switch (cond.ComparisionOperator) {
                case FilterComparisionOperator.Contains:
                    output += "LIKE '*" + cond.CompareValue.ToString() + "*'";
                    break;
                case FilterComparisionOperator.Custom:
                    throw new NotImplementedException();
                    break;
                case FilterComparisionOperator.DoesNotContain:
                    output += "NOT LIKE '*" + cond.CompareValue.ToString() + "*'";
                    break;
                case FilterComparisionOperator.DoesNotEndWith:
                    output += "NOT LIKE '*" + cond.CompareValue.ToString() + "'";
                    break;
                case FilterComparisionOperator.DoesNotMatch:
                    throw new NotImplementedException();
                    break;
                case FilterComparisionOperator.DoesNotStartWith:
                    output += "NOT LIKE '" + cond.CompareValue.ToString() + "*'";
                    break;
                case FilterComparisionOperator.EndsWith:
                    output += "LIKE '*" + cond.CompareValue.ToString() + "'";
                    break;
                case FilterComparisionOperator.Equals:
                    output += "= " + cond.CompareValue.ToString();
                    break;
                case FilterComparisionOperator.GreaterThan:
                    output += "> " + cond.CompareValue.ToString();
                    break;
                case FilterComparisionOperator.GreaterThanOrEqualTo:
                    output += ">= " + cond.CompareValue.ToString();
                    break;
                case FilterComparisionOperator.LessThan:
                    output += "<" + cond.CompareValue.ToString();
                    break;
                case FilterComparisionOperator.LessThanOrEqualTo:
                    output += "<=" + cond.CompareValue.ToString();
                    break;
                case FilterComparisionOperator.Like:
                    output += "LIKE '" + cond.CompareValue.ToString() + "'";
                    break;
                case FilterComparisionOperator.Match:
                    throw new NotImplementedException();
                    break;
                case FilterComparisionOperator.NotEquals:
                    output += "<> " + cond.CompareValue.ToString();
                    break;
                case FilterComparisionOperator.NotLike:
                    output += "NOT LIKE '" + cond.CompareValue.ToString() + "'";
                    break;
                case FilterComparisionOperator.StartsWith:
                    output += "LIKE '" + cond.CompareValue.ToString() + "*'";
                    break;
            }
            return output;
        }

    }


}
